/* Author: Derek O Reilly, Dundalk Institute of Technology, Ireland. */


class TeddyBodyPart
{
    /* Each gameObject MUST have a constructor() and a render() method.        */
    /* If the object animates, then it must also have an updateState() method. */

    constructor(scale, sourceImage, width, height, radiants, centreOfRotationX, centreOfRotationY, hotSpotX, hotSpotY, hotSpotRadius)
    {
        // There needs to be one bodyPart for each seperate part of the body
        // For example, the left arm consists of 3 bodyParts: LEFT_ARM_UPPER, LEFT_ARM_LOWER and LEFT_HAND

        /* These variables depend on the object */
        this.scale = scale; // a scale of 1.0 means this whole Teddy is the size of the canvas.
        this.bodyPartImageFile = sourceImage; // the image file that contains this bodyPart

        this.width = width * scale;        // width and height of the bodyPart in canvas coordinates
        this.height = height * scale;
        this.radiants = radiants;      // the angle of rotation (in radiants)
        this.centreOfRotationX = centreOfRotationX * scale;  // the point around which rotations occur
        this.centreOfRotationY = centreOfRotationY * scale;

        this.hotSpotX = hotSpotX * scale;             // hot spot for identifying where the user can drag the mouse to 
        this.hotSpotY = hotSpotY * scale;             // rotate this bodyPart
        this.hotSpotRadius = hotSpotRadius * scale;   // size of the hotspot

        this.children = [];
        this.numberOfChildren = 0;      // used to keep track of the number of children in the this.children array above.

        this.hotSpotIsDrawn = false;  // by default, do not draw the hotspot on the screen

        /* this.offscreenBodyPartCanvasCtx will be used to hold the rotated bodyPart */
        /* When a bodyPart rotates, it will also cause its children to rotate */
        this.offscreenBodyPartCanvas = document.createElement('canvas');
        this.offscreenBodyPartCanvasCtx = this.offscreenBodyPartCanvas.getContext('2d');
        this.offscreenBodyPartCanvas.width = canvas.width;
        this.offscreenBodyPartCanvas.height = canvas.height;
    }

    render()
    {
        // draw rotated bodyPart onto its offscreen canvas     
        this.offscreenBodyPartCanvasCtx.clearRect(0, 0, canvas.width, canvas.height);
        this.rotateOffscreenCanvas(this.radiants, this.parentX, this.parentY);
        this.offscreenBodyPartCanvasCtx.drawImage(this.bodyPartImageFile, this.parentX - this.centreOfRotationX, this.parentY - this.centreOfRotationY, this.width, this.height);
        this.rotateOffscreenCanvas(-this.radiants, this.parentX, this.parentY);

        // if it is enabled, then draw the the hotspot on the offscreen canvas
        if (this.hotSpotIsDrawn === true)
        {
            this.drawHotSpotOnOffscreenCanvas();
        }

        // draw the offscreen canvas onto the main canvas
        ctx.drawImage(this.offscreenBodyPartCanvas, 0, 0, canvas.width, canvas.height);
    }

    setChildrenPositions()
    {
        // This method adjusts the postion of this bodyPart's children to take account of this bodyPart's postion.
        // To achieve the adjustment of position, this method sets the 'this.parentX', 'this.parentY' and 'this.radiant' variables of this bodyPart's children

        // update children to account for the rotation of this bodyPart 
        for (var i = 0; i < this.numberOfChildren; i++)
        {
            if (this.children[i] !== undefined)
            {
                // rotate each child to account for the rotation of this bodyPart
                this.children[i].updateRotatedParentXandY(this.parentX, this.parentY, this.radiants, this.centreOfRotationX, this.centreOfRotationY);

                // recursively rotate each child bodyPart about its own centre of rotation
                this.children[i].setChildrenPositions();
            }
        }
    }

    updateRotatedParentXandY(centreOfRotationX, centreOfRotationY, radiants, offsetX, offsetY)
    {
        // rotate this bodyPart to account for rotations of this bodyPart's parent
        this.parentX = this.originalParentX;
        this.parentY = this.originalParentY;
        this.parentX -= offsetX;
        this.parentY -= offsetY;

        let newX = this.parentX * Math.cos(radiants) - this.parentY * Math.sin(radiants);
        this.parentY = this.parentX * Math.sin(radiants) + this.parentY * Math.cos(radiants);
        this.parentX = newX;

        this.parentX += centreOfRotationX;
        this.parentY += centreOfRotationY;
    }
    
    rotateOffscreenCanvas(radiants, centreX, centreY)
    {
        // this is a helper method for the method 'rotate()'
        // rotate the offscreen canvas of this bodyPart around this the given centre of rotation    
        this.offscreenBodyPartCanvasCtx.translate(centreX, centreY);
        this.offscreenBodyPartCanvasCtx.rotate(radiants);
        this.offscreenBodyPartCanvasCtx.translate(-centreX, -centreY);
    }

    changeRotationValue(mouseX, mouseY, degrees)
    {
        // depending on the position of the mouse relative to the position of this.parentX and this.parentY, add or subtract degrees from this.radiants
        if (this.isRotatingClockwise(this.parentX, this.parentY, mouseX, mouseY) === true)
        {
            this.radiants += Math.radiants(degrees);
        }
        else
        {
            this.radiants -= Math.radiants(degrees);
        }
    }

    isRotatingClockwise(centerX, centerY, x, y)
    {
        // return true if the mouse is dragging in a clockwise direction, else return false
        if (Math.abs(x - oldMouseX) >= Math.abs(y - oldMouseY))
        {
            // mouse is being moved primarily along the x-axis
            if ((x < oldMouseX))  // mouse is moving to the left
            {
                if (y < centerY)     // mouse is above the image
                {
                    return false;
                }
                else   // mouse is below the image
                {
                    return true;
                }
            }
            else  // mouse is moving to the right
            {
                if (y < centerY)   // mouse is to the left of the image
                {
                    return true;
                }
                else  // mouse is to the right of the image
                {
                    return false;
                }
            }
        }
        else // mouse is being moved primarily along the y-axis
        {
            // mouse is being moved primarily along the y-axis
            if (y < oldMouseY)  // mouse is moving to the left
            {
                if (x < centerX)     // mouse is above the image
                {
                    return true;
                }
                else   // mouse is below the image
                {
                    return false;
                }
            }
            else  // mouse is moving to the right
            {
                if (x < centerX)   // mouse is to the left of the image
                {
                    return false;
                }
                else  // mouse is to the right of the image
                {
                    return true;
                }
            }
        }
    }

    isHotSpot(canvasX, canvasY)
    {
        // return true if the canvasX, canvasY is in this bodyPart's hotspot
        var hotSpotX = this.parentX - this.centreOfRotationX + this.hotSpotX;
        var hotSpotY = this.parentY - this.centreOfRotationY + this.hotSpotY;

        // rotate the hotspot around the parentX and parentY
        hotSpotX -= this.parentX;
        hotSpotY -= this.parentY;

        let newX = hotSpotX * Math.cos(this.radiants) - hotSpotY * Math.sin(this.radiants);
        hotSpotY = hotSpotX * Math.sin(this.radiants) + hotSpotY * Math.cos(this.radiants);
        hotSpotX = newX;

        hotSpotX += this.parentX;
        hotSpotY += this.parentY;

        if ((canvasX > hotSpotX - this.hotSpotRadius) && (canvasX < hotSpotX + this.hotSpotRadius) &&
                (canvasY > hotSpotY - this.hotSpotRadius) && (canvasY < hotSpotY + this.hotSpotRadius))
        {
            return true;
        }
        else
        {
            return false;
        }
    }

    drawHotSpotOnOffscreenCanvas()
    {
        // draw a bodyPart's hotspot on the offscreenBodyPartCanvas
        if (this.hotSpotIsDrawn === false)
        {
            return;
        }

        var hotSpotX = this.parentX - this.centreOfRotationX + this.hotSpotX;
        var hotSpotY = this.parentY - this.centreOfRotationY + this.hotSpotY;

        // rotate the hotspot around the parentX and parentY
        hotSpotX -= this.parentX;
        hotSpotY -= this.parentY;

        let newX = hotSpotX * Math.cos(this.radiants) - hotSpotY * Math.sin(this.radiants);
        hotSpotY = hotSpotX * Math.sin(this.radiants) + hotSpotY * Math.cos(this.radiants);
        hotSpotX = newX;

        hotSpotX += this.parentX;
        hotSpotY += this.parentY;

        // draw the hotspot
        this.offscreenBodyPartCanvasCtx.globalAlpha = 0.15;
        this.offscreenBodyPartCanvasCtx.beginPath();
        this.offscreenBodyPartCanvasCtx.fillStyle = "blue";
        this.offscreenBodyPartCanvasCtx.arc(hotSpotX, hotSpotY, this.hotSpotRadius, 0, Math.PI * 2);
        this.offscreenBodyPartCanvasCtx.fill();
        this.offscreenBodyPartCanvasCtx.closePath();
        this.offscreenBodyPartCanvasCtx.globalAlpha = 1;
    }

    setChild(child, parentX, parentY)
    {
        // set a direct child of this bodyPart
        // For example, LEFT_ARM_UPPER has one direct child, which is LEFT_ARM_LOWER
        this.children[this.numberOfChildren] = child;
        child.setInitialParentXandY(parentX * this.scale, parentY * this.scale);
        this.numberOfChildren++;
    }

    setInitialParentXandY(newParentX, newParentY)
    {
        // this method ties the parent's coordinates to the child's centreOfRotation coordinates
        // The parent's coordinates are the x,y values of the joint on the parent where
        // the child's rotation x,y coordinates will attach to
        this.parentX = newParentX;
        this.parentY = newParentY;
        this.originalParentX = this.parentX;
        this.originalParentY = this.parentY;
    }
    
    setDrawHotSpot(state)
    {
        // Not strictly necessary, but might be useful as a guide to showing younger users where to drag
        this.hotSpotIsDrawn = state;
    }

    getDegrees()
    {
        // return the rotation angle as degrees
        return this.radiants * 180 / Math.PI;
    }

    setRadiants(newDegrees)
    {
        // set the rotation angle as radiants
        // note that the input parameter, newDegrees, is given in degrees
        this.radiants = Math.radiants(newDegrees);
    }

    setDrawHotSpot(state)
    {
        this.hotSpotIsDrawn = state;
    }
}